id ICON "hr-icon.ico"

1 VERSIONINFO
FILEVERSION     10,0,0,14
PRODUCTVERSION  10,0,0,14
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
      VALUE "CompanyName", "Zeno Rogue"
      VALUE "FileDescription", "A roguelike in non-euclidean space"
      VALUE "FileVersion", "A10.0n"
      VALUE "InternalName", "hyper"
      VALUE "LegalCopyright", "Zeno Rogue"
      VALUE "OriginalFilename", "hyper.exe"
      VALUE "ProductName", "HyperRogue"
      VALUE "ProductVersion", "10.0n"
    END
  END

  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x409, 1252
  END
END
